@@ -265,17 +265,43 @@ start_server {tags {"scripting"}} {
         run_script {return redis.pcall('wait','1','0')} 0
     } {0}
 
-    test {EVAL - Scripts can't run XREAD and XREADGROUP with BLOCK option} {
+    test {EVAL - Scripts do not block on waitaof} {
+        r config set appendonly no
+        run_script {return redis.pcall('waitaof','0','1','0')} 0
+    } {0 0}
+
+    test {EVAL - Scripts do not block on XREAD with BLOCK option} {
         r del s
         r xgroup create s g $ MKSTREAM
         set res [run_script {return redis.pcall('xread','STREAMS','s','$')} 1 s]
         assert {$res eq {}}
-        assert_error "*xread command is not allowed with BLOCK option from scripts" {run_script {return redis.pcall('xread','BLOCK',0,'STREAMS','s','$')} 1 s}
+        run_script {return redis.pcall('xread','BLOCK',0,'STREAMS','s','$')} 1 s
+    } {}
+
+    test {EVAL - Scripts do not block on XREADGROUP with BLOCK option} {
         set res [run_script {return redis.pcall('xreadgroup','group','g','c','STREAMS','s','>')} 1 s]
         assert {$res eq {}}
-        assert_error "*xreadgroup command is not allowed with BLOCK option from scripts" {run_script {return redis.pcall('xreadgroup','group','g','c','BLOCK',0,'STREAMS','s','>')} 1 s}
+        run_script {return redis.pcall('xreadgroup','group','g','c','BLOCK',0,'STREAMS','s','>')} 1 s
+    } {}
+
+    test {EVAL - Scripts do not block on XREAD with BLOCK option -- non empty stream} {
+        r XADD s * a 1
+        set res [run_script {return redis.pcall('xread','BLOCK',0,'STREAMS','s','$')} 1 s]
+        assert {$res eq {}}
+
+        set res [run_script {return redis.pcall('xread','BLOCK',0,'STREAMS','s','0-0')} 1 s]
+        assert {[lrange [lindex $res 0 1 0 1] 0 1] eq {a 1}}
     }
 
+    test {EVAL - Scripts do not block on XREADGROUP with BLOCK option -- non empty stream} {
+        r XADD s * b 2
+        set res [
+            run_script {return redis.pcall('xreadgroup','group','g','c','BLOCK',0,'STREAMS','s','>')} 1 s
+        ]
+        assert {[llength [lindex $res 0 1]] == 2}
+        lindex $res 0 1 0 1
+    } {a 1}
+
     test {EVAL - Scripts can run non-deterministic commands} {
         set e {}
         catch {
@@ -315,44 +341,12 @@ start_server {tags {"scripting"}} {
         set e
     } {*against a key*}
 
-    test {EVAL - Test table unpack with invalid indexes} {
-        catch {run_script { return {unpack({1,2,3}, -2, 2147483647)} } 0} e
-        assert_match {*too many results to unpack*} $e
-        catch {run_script { return {unpack({1,2,3}, 0, 2147483647)} } 0} e
-        assert_match {*too many results to unpack*} $e
-        catch {run_script { return {unpack({1,2,3}, -2147483648, -2)} } 0} e
-        assert_match {*too many results to unpack*} $e
-        set res [run_script { return {unpack({1,2,3}, -1, -2)} } 0]
-        assert_match {} $res
-        set res [run_script { return {unpack({1,2,3}, 1, -1)} } 0]
-        assert_match {} $res
-
-        # unpack with range -1 to 5, verify nil indexes
-        set res [run_script {
-             local function unpack_to_list(t, i, j)
-               local n, v = select('#', unpack(t, i, j)), {unpack(t, i, j)}
-               for i = 1, n do v[i] = v[i] or '_NIL_' end
-               v.n = n
-               return v
-             end
-
-            return unpack_to_list({1,2,3}, -1, 5)
-        } 0]
-        assert_match {_NIL_ _NIL_ 1 2 3 _NIL_ _NIL_} $res
-
-        # unpack with negative range, verify nil indexes
-        set res [run_script {
-             local function unpack_to_list(t, i, j)
-               local n, v = select('#', unpack(t, i, j)), {unpack(t, i, j)}
-               for i = 1, n do v[i] = v[i] or '_NIL_' end
-               v.n = n
-               return v
-             end
-
-            return unpack_to_list({1,2,3}, -2147483648, -2147483646)
-        } 0]
-        assert_match {_NIL_ _NIL_ _NIL_} $res
-    } {}
+    test {EVAL - JSON string encoding a string larger than 2GB} {
+        run_script {
+            local s = string.rep("a", 1024 * 1024 * 1024)
+            return #cjson.encode(s..s..s)
+        } 0
+    } {3221225474} {large-memory} ;# length includes two double quotes at both ends
 
     test {EVAL - JSON numeric decoding} {
         # We must return the table as a string because otherwise
@@ -583,13 +577,19 @@ start_server {tags {"scripting"}} {
     # script command is only relevant for is_eval Lua
     test {SCRIPTING FLUSH - is able to clear the scripts cache?} {
         r set mykey myval
+
+        r script load {return redis.call('get',KEYS[1])}
         set v [r evalsha fd758d1589d044dd850a6f05d52f2eefd27f033f 1 mykey]
         assert_equal $v myval
-        set e ""
         r script flush
-        catch {r evalsha fd758d1589d044dd850a6f05d52f2eefd27f033f 1 mykey} e
-        set e
-    } {NOSCRIPT*}
+        assert_error {NOSCRIPT*} {r evalsha fd758d1589d044dd850a6f05d52f2eefd27f033f 1 mykey}
+
+        r eval {return redis.call('get',KEYS[1])} 1 mykey
+        set v [r evalsha fd758d1589d044dd850a6f05d52f2eefd27f033f 1 mykey]
+        assert_equal $v myval
+        r script flush
+        assert_error {NOSCRIPT*} {r evalsha fd758d1589d044dd850a6f05d52f2eefd27f033f 1 mykey}
+    }
 
     test {SCRIPTING FLUSH ASYNC} {
         for {set j 0} {$j < 100} {incr j} {
@@ -635,6 +635,52 @@ start_server {tags {"scripting"}} {
              [run_script {return redis.sha1hex('Pizza & Mandolino')} 0]
     } {da39a3ee5e6b4b0d3255bfef95601890afd80709 74822d82031af7493c20eefa13bd07ec4fada82f}
 
+    test "Measures elapsed time os.clock()" {
+        set escaped [run_script {
+            local start = os.clock()
+            while os.clock() - start < 1 do end
+            return {double = os.clock() - start}
+        } 0]
+        assert_morethan_equal $escaped 1 ;# 1 second
+    }
+
+    test "Prohibit dangerous lua methods in sandbox" {
+        assert_equal "" [run_script {
+            local allowed_methods = {"clock"}
+            -- Find a value from a tuple and return the position.
+            local indexOf = function(tuple, value)
+                for i, v in ipairs(tuple) do
+                    if v == value then return i end
+                end
+                return nil
+            end
+            -- Check for disallowed methods and verify all allowed methods exist.
+            -- If an allowed method is found, it's removed from 'allowed_methods'.
+            -- If 'allowed_methods' is empty at the end, all allowed methods were found.
+            for key, value in pairs(os) do
+                local index = indexOf(allowed_methods, key)
+                if index == nil or type(value) ~= "function" then
+                    return "Disallowed "..type(value)..":"..key
+                end
+                table.remove(allowed_methods, index)
+            end
+            if #allowed_methods ~= 0 then
+                return "Expected method not found: "..table.concat(allowed_methods, ",")
+            end
+            return ""
+        } 0]
+    }
+
+    test "Verify execution of prohibit dangerous Lua methods will fail" {
+        assert_error {ERR *attempt to call field 'execute'*} {run_script {os.execute()} 0}
+        assert_error {ERR *attempt to call field 'exit'*} {run_script {os.exit()} 0}
+        assert_error {ERR *attempt to call field 'getenv'*} {run_script {os.getenv()} 0}
+        assert_error {ERR *attempt to call field 'remove'*} {run_script {os.remove()} 0}
+        assert_error {ERR *attempt to call field 'rename'*} {run_script {os.rename()} 0}
+        assert_error {ERR *attempt to call field 'setlocale'*} {run_script {os.setlocale()} 0}
+        assert_error {ERR *attempt to call field 'tmpname'*} {run_script {os.tmpname()} 0}
+    }
+
     test {Globals protection reading an undeclared global variable} {
         catch {run_script {return a} 0} e
         set e
@@ -645,12 +691,6 @@ start_server {tags {"scripting"}} {
         set e
     } {ERR *Attempt to modify a readonly table*}
 
-    test {lua bit.tohex bug} {
-        set res [run_script {return bit.tohex(65535, -2147483648)} 0]
-        r ping
-        set res
-    } {0000FFFF}
-
     test {Test an example script DECR_IF_GT} {
         set decr_if_gt {
             local current
@@ -1039,27 +1079,6 @@ start_server {tags {"scripting"}} {
         set _ $e
     } {*Attempt to modify a readonly table*}
 
-    test "Try trick readonly table on basic types metatable" {
-        # Run the following scripts for basic types. Either getmetatable()
-        # should return nil or the metatable must be readonly.
-        set scripts {
-            {getmetatable(nil).__index = function() return 1 end}
-            {getmetatable('').__index = function() return 1 end}
-            {getmetatable(123.222).__index = function() return 1 end}
-            {getmetatable(true).__index = function() return 1 end}
-            {getmetatable(function() return 1 end).__index = function() return 1 end}
-            {getmetatable(coroutine.create(function() return 1 end)).__index = function() return 1 end}
-        }
-
-        foreach code $scripts {
-            catch {run_script $code 0} e
-            assert {
-                [string match "*attempt to index a nil value script*" $e] ||
-                [string match "*Attempt to modify a readonly table*" $e]
-            }
-        }
-    }
-
     test "Test loadfile are not available" {
         catch {
             run_script {
@@ -1088,102 +1107,6 @@ start_server {tags {"scripting"}} {
     } {*Script attempted to access nonexistent global variable 'print'*}
 }
 
-# start a new server to test the large-memory tests
-start_server {tags {"scripting external:skip large-memory"}} {
-
-    test {EVAL - JSON string encoding a string larger than 2GB} {
-        run_script {
-            local s = string.rep("a", 1024 * 1024 * 1024)
-            return #cjson.encode(s..s..s)
-        } 0
-    } {3221225474} ;# length includes two double quotes at both ends
-
-    test {EVAL - Test long escape sequences for strings} {
-        run_script {
-            -- Generate 1gb '==...==' separator
-            local s = string.rep('=', 1024 * 1024)
-            local t = {} for i=1,1024 do t[i] = s end
-            local sep = table.concat(t)
-            collectgarbage('collect')
-
-            local code = table.concat({'return [',sep,'[x]',sep,']'})
-            collectgarbage('collect')
-
-            -- Load the code and run it. Script will return the string length.
-            -- Escape sequence: [=....=[ to ]=...=] will be ignored
-            -- Actual string is a single character: 'x'. Script will return 1
-            local func = loadstring(code)
-            return #func()
-        } 0
-    } {1}
-
-    test {EVAL - Lua can parse string with too many new lines} {
-        # Create a long string consisting only of newline characters. When Lua
-        # fails to parse a string, it typically includes a snippet like
-        # "... near ..." in the error message to indicate the last recognizable
-        # token. In this test, since the input contains only newlines, there
-        # should be no identifiable token, so the error message should contain
-        # only the actual error, without a near clause.
-
-        run_script {
-           local s = string.rep('\n', 1024 * 1024)
-           local t = {} for i=1,2048 do t[#t+1] = s end
-           local lines = table.concat(t)
-           local fn, err = loadstring(lines)
-           return err
-        } 0
-    } {*chunk has too many lines}
-}
-
-# Start a new server to test lua-enable-deprecated-api config
-foreach enabled {no yes} {
-start_server [subst {tags {"scripting external:skip"} overrides {lua-enable-deprecated-api $enabled}}] {
-    test "Test setfenv availability lua-enable-deprecated-api=$enabled" {
-        catch {
-            run_script {
-                local f = function() return 1 end
-                setfenv(f, {})
-                return 0
-            } 0
-        } e
-        if {$enabled} {
-            assert_equal $e 0
-        } else {
-            assert_match {*Script attempted to access nonexistent global variable 'setfenv'*} $e
-        }
-    }
-
-    test "Test getfenv availability lua-enable-deprecated-api=$enabled" {
-        catch {
-            run_script {
-                local f = function() return 1 end
-                getfenv(f)
-                return 0
-            } 0
-        } e
-        if {$enabled} {
-            assert_equal $e 0
-        } else {
-            assert_match {*Script attempted to access nonexistent global variable 'getfenv'*} $e
-        }
-    }
-
-    test "Test newproxy availability lua-enable-deprecated-api=$enabled" {
-        catch {
-            run_script {
-                getmetatable(newproxy(true)).__gc = function() return 1 end
-                return 0
-            } 0
-        } e
-        if {$enabled} {
-            assert_equal $e 0
-        } else {
-            assert_match {*Script attempted to access nonexistent global variable 'newproxy'*} $e
-        }
-    }
-}
-}
-
 # Start a new server since the last test in this stanza will kill the
 # instance at all.
 start_server {tags {"scripting"}} {
@@ -1598,6 +1521,93 @@ start_server {tags {"scripting needs:debug external:skip"}} {
         assert_equal [r ping] {PONG}
     }
 }
+
+start_server {tags {"scripting external:skip"}} {
+    test {Lua scripts eviction does not generate many scripts} {
+        r script flush
+        r config resetstat
+
+        # "return 1" sha is: e0e1f9fabfc9d4800c877a703b823ac0578ff8db
+        # "return 500" sha is: 98fe65896b61b785c5ed328a5a0a1421f4f1490c
+        for {set j 1} {$j <= 250} {incr j} {
+            r eval "return $j" 0
+        }
+        for {set j 251} {$j <= 500} {incr j} {
+            r eval_ro "return $j" 0
+        }
+        assert_equal [s number_of_cached_scripts] 500
+        assert_equal 1 [r evalsha e0e1f9fabfc9d4800c877a703b823ac0578ff8db 0]
+        assert_equal 1 [r evalsha_ro e0e1f9fabfc9d4800c877a703b823ac0578ff8db 0]
+        assert_equal 500 [r evalsha 98fe65896b61b785c5ed328a5a0a1421f4f1490c 0]
+        assert_equal 500 [r evalsha_ro 98fe65896b61b785c5ed328a5a0a1421f4f1490c 0]
+
+        # Scripts between "return 1" and "return 500" are evicted
+        for {set j 501} {$j <= 750} {incr j} {
+            r eval "return $j" 0
+        }
+        for {set j 751} {$j <= 1000} {incr j} {
+            r eval "return $j" 0
+        }
+        assert_error {NOSCRIPT*} {r evalsha e0e1f9fabfc9d4800c877a703b823ac0578ff8db 0}
+        assert_error {NOSCRIPT*} {r evalsha_ro e0e1f9fabfc9d4800c877a703b823ac0578ff8db 0}
+        assert_error {NOSCRIPT*} {r evalsha 98fe65896b61b785c5ed328a5a0a1421f4f1490c 0}
+        assert_error {NOSCRIPT*} {r evalsha_ro 98fe65896b61b785c5ed328a5a0a1421f4f1490c 0}
+
+        assert_equal [s evicted_scripts] 500
+        assert_equal [s number_of_cached_scripts] 500
+    }
+
+    test {Lua scripts eviction is plain LRU} {
+        r script flush
+        r config resetstat
+
+        # "return 1" sha is: e0e1f9fabfc9d4800c877a703b823ac0578ff8db
+        # "return 2" sha is: 7f923f79fe76194c868d7e1d0820de36700eb649
+        # "return 3" sha is: 09d3822de862f46d784e6a36848b4f0736dda47a
+        # "return 500" sha is: 98fe65896b61b785c5ed328a5a0a1421f4f1490c
+        # "return 1000" sha is: 94f1a7bc9f985a1a1d5a826a85579137d9d840c8
+        for {set j 1} {$j <= 500} {incr j} {
+            r eval "return $j" 0
+        }
+
+        # Call "return 1" to move it to the tail.
+        r eval "return 1" 0
+        # Call "return 2" to move it to the tail.
+        r evalsha 7f923f79fe76194c868d7e1d0820de36700eb649 0
+        # Create a new script, "return 3" will be evicted.
+        r eval "return 1000" 0
+        # "return 1" is ok since it was moved to tail.
+        assert_equal 1 [r evalsha e0e1f9fabfc9d4800c877a703b823ac0578ff8db 0]
+        # "return 2" is ok since it was moved to tail.
+        assert_equal 1 [r evalsha e0e1f9fabfc9d4800c877a703b823ac0578ff8db 0]
+        # "return 3" was evicted.
+        assert_error {NOSCRIPT*} {r evalsha 09d3822de862f46d784e6a36848b4f0736dda47a 0}
+        # Others are ok.
+        assert_equal 500 [r evalsha 98fe65896b61b785c5ed328a5a0a1421f4f1490c 0]
+        assert_equal 1000 [r evalsha 94f1a7bc9f985a1a1d5a826a85579137d9d840c8 0]
+
+        assert_equal [s evicted_scripts] 1
+        assert_equal [s number_of_cached_scripts] 500
+    }
+
+    test {Lua scripts eviction does not affect script load} {
+        r script flush
+        r config resetstat
+
+        set num [randomRange 500 1000]
+        for {set j 1} {$j <= $num} {incr j} {
+            r script load "return $j"
+            r eval "return 'str_$j'" 0
+        }
+        set evicted [s evicted_scripts]
+        set cached [s number_of_cached_scripts]
+        # evicted = num eval scripts - 500 eval scripts
+        assert_equal $evicted [expr $num-500]
+        # cached = num load scripts + 500 eval scripts
+        assert_equal $cached [expr $num+500]
+    }
+}
+
 } ;# is_eval
 
 start_server {tags {"scripting needs:debug"}} {
@@ -1866,6 +1876,27 @@ start_server {tags {"scripting needs:debug"}} {
 
     r debug set-disable-deny-scripts 0
 }
+
+start_server {tags {"scripting"}} {
+    test "Verify Lua performs GC correctly after script loading" {
+        set dummy_script "--[string repeat x 10]\nreturn "
+        set n 50000
+        for {set i 0} {$i < $n} {incr i} {
+            set script "$dummy_script[format "%06d" $i]"
+            if {$is_eval} {
+                r script load $script
+            } else {
+                r function load "#!lua name=test$i\nredis.register_function('test$i', function(KEYS, ARGV)\n $script \nend)"
+            }
+        }
+
+        if {$is_eval} {
+            assert_lessthan [s used_memory_lua] 17500000
+        } else {
+            assert_lessthan [s used_memory_vm_functions] 14500000
+        }
+    }
+}
 } ;# foreach is_eval
 
 
